/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "parse_stage.hh"
// #include "chan_list.hh"
#include "buffer.hh"
#include "sysexcept.hh"
#include <iostream>
#include <signal.h>
#include <cstdlib>

#include "dbase.hh"

using namespace sends;
using namespace std;

//======================================  Construct parse stage
parse_stage::parse_stage(int nThread, int nTrans)
    : seda_stage(nThread, nTrans), resource(nThread, parse_buffer_length),
    timeout(2), quit_count(0)
{
}

//======================================  Destroy parse stage
parse_stage::~parse_stage(void) {
}

//======================================  Return stage name
const char*
parse_stage::stage_name(void) const {
    return "parser";
}

//======================================  Skip spaces
inline void
skip_spaces(const char*& ptr, const char* end) {
    while (*ptr == ' ' && ptr < end) ptr++;
}

//======================================  Get the channel type
inline chan_type
get_chan_type(const transaction& t) {
    index_type nChan = t.getNChan();
    if (!nChan) return chan_type::kUnknown;
    chan_type r = Channel_List[t.chan(0).index()].type();
    for (index_type i=1; i<nChan; ++i) {
	if (r != Channel_List[t.chan(i).index()].type()) {
	    return chan_type::kUnknown;
	}
    }
    return r;
}

//======================================  Parse client command.
void
parse_stage::filter(transaction* trans) {
    buffer* buf = 0;
    int     len = 0;

    //----------------------------------Get a command line.
    trans->set_stage(200);
    trans->set_state(transaction::ws_initial);
    trans->setTime(0, 0, 0);
    try {
	socket_api& s(trans->ref_socket());

	//------------------------------  Wait for a request
	int dok = s.wait_data(timeout);
	if (dok <= 0) {
	    string msg = "Data wait timed out in ";
	    msg += s.where();
	    throw system_error(msg, s.sys_errno());
	}

	//------------------------------  Get a preallocated buffer
        buf = resource.get_buffer();

	//------------------------------  Get the command line
        len = trans->read(*buf);
        if (len <= 0) {
	    string msg = "buffer read failed in ";
	    msg += s.where();
	    throw system_error(msg, s.sys_errno());
	}

    //----------------------------------  Exception - dump the transaction
    } catch (exception& e) {
	trans->set_stage(210);
	cancel(trans, e.what());
	if (buf) buf->free();
	return;
    }

    //----------------------------------  Parse the command line
    trans->set_stage(210);
    try {
	//------------------------------  Get the command name
	const char* ptr = buf->data();
	const char* end = ptr + len;
	debug_msg(dlvl_transaction, trans->ident(), string("command: ")+ptr);
	parse_command(ptr, end, trans);

	//------------------------------  Parse commands
	const int maxarg=8;
	long iarg[maxarg];
	int narg;
	switch (trans->command()) {

	//------------------------------  Parse get-data command
	case transaction::kGetData:
	    narg = parse_times(ptr, end, 3, iarg);
	    if (narg < 2) {
		throw runtime_error("Insufficient numeric arguments");
	    } else if (narg == 2 && iarg[0] < iarg[1]) {
		trans->setTime(iarg[0], iarg[1]);
	    } else if (narg == 2) {
		trans->setTime(iarg[0], iarg[0]+iarg[1]);
	    } else if (iarg[0] < iarg[1] && 
		       iarg[2] < 2000000000 &&
		       iarg[2] >= 0) {
		trans->setTime(iarg[0], iarg[1], iarg[2]);
	    } else {
		throw runtime_error("Invalid time parameters");
	    }
	    parse_chantype(ptr, end, trans);
	    parse_channels(ptr, end, trans);

	    //--------------------------  Online compatibility
	    //  If all the input channels are online, make this into an 
	    //  request using offline protocol.
	    if (get_chan_type(*trans) == chan_type::kOnline) {
		debug_msg(dlvl_transaction, trans->ident(), 
			  "Command switched to get-online-data.");
		trans->setCommand("get-online-data");
	    }
	    break;

	//------------------------------  Parse get-data command
	case transaction::kGetOnlineData:
	    narg = parse_times(ptr, end, 2, iarg);
	    if (narg == 1) {
		trans->setTime(0, 0, iarg[0]);
	    } else if (narg == 2) {
		trans->setTime(0, iarg[0], iarg[1]);
	    }
	    trans->setDefType(chan_type::kUnknown);
	    parse_channels(ptr, end, trans);
	    break;

	//------------------------------  Parse get-source-list command
	case transaction::kGetSourceList:
	case transaction::kGetSourceData:
	    narg = parse_times(ptr, end, 1, iarg);
	    if (narg < 1) trans->setTime(0, 0);
	    else          trans->setTime(iarg[0], 0);
	    trans->setDefType(chan_type::kUnknown);
	    parse_channels(ptr, end, trans);
	    break;

	//------------------------------  Parse get-channels command
	case transaction::kGetChannels:
	case transaction::kCountChannels:
	case transaction::kChannelCRC:
	    narg = parse_times(ptr, end, 1, iarg);
	    if (narg < 1) trans->setTime(0, 0);
	    else          trans->setTime(iarg[0], 0);
	    parse_chantype(ptr, end, trans);
	    break;

	//------------------------------  Parse set epoch
	case transaction::kSetEpoch:
	    parse_epoch(ptr, end, trans);
	    break;

	//------------------------------  Commands with no arguments
	case transaction::kGetServerStatus:
	    dump_server(cout);
	    break;

	//------------------------------  Commands with no arguments
	case transaction::kSegmentStatus:
	case transaction::kServerProtocolVersion:
	case transaction::kServerProtocolRevision:
	case transaction::kListEpochs:
	case transaction::kQuit:
	    break;
	default:
	    throw runtime_error("Command syntax error");
	}

	//------------------------------  Check semicolon.
	skip_spaces(ptr, end);
	if (*ptr != ';') throw runtime_error("No semicolon");

	//------------------------------  Generate appropriate response and
	//                                send transaction to the next stage.
	trans->set_stage(210);
	switch (trans->command()) {

	//------------------------------  Handle get-data commands
	case transaction::kGetData:
	case transaction::kGetOnlineData:
	    //--------------------------  Check requirements aren't excessive
	    if (!trans->test_limits()) {
		throw runtime_error("Excessive resource requirements");
	    }
	    //--------------------------  Defer response to later stage
	    put_output(trans);
	    break;

	//------------------------------  Route immediate commands to alternate
        case transaction::kGetServerStatus:
	case transaction::kSetEpoch:
	    trans->put_response(DAQD_OK);
	    put_alternate(trans);
	    break;

        //------------------------------  Pass along commands handled by write
        case transaction::kGetChannels:
        case transaction::kCountChannels:
        case transaction::kChannelCRC:
        case transaction::kSegmentStatus:
        case transaction::kGetSourceList:
        case transaction::kGetSourceData:
	case transaction::kServerProtocolVersion:
	case transaction::kServerProtocolRevision:
	case transaction::kListEpochs:
            trans->put_response(DAQD_OK);
            put_output(trans);
            break;

	//------------------------------  No response to Quit command
	case transaction::kQuit:
	    quit_count++;
	    trans->release();
	    break;

	//------------------------------  Most commands go to the normal output
	default:
	    throw runtime_error("Command syntax error");
	}

    //----------------------------------  Exception - dump the transaction
    } catch (exception& e) {
	trans->set_stage(220);
	if (!strncmp(e.what(), "Unknown or unavailable channel", 30)) {
	    trans->put_response(DAQD_INVALID_CHANNEL_NAME);
	} else if (!strncmp(e.what(), "Excessive resource requirements", 31)) {
	    trans->put_response(DAQD_TOO_MANY_CHANNELS);
	} else {
	    trans->put_response(DAQD_COMMAND_SYNTAX);
	}
	reject(trans, e.what());
    }

    //----------------------------------  Release buffer
    if (buf) {
        buf->free();
    }
}

//======================================  Parse channel list.
void 
parse_stage::parse_channels(const char* &ptr, const char* end, transaction* t) {

    //----------------------------------  Get start of list
    skip_spaces(ptr, end);
    if (*ptr != '{') throw runtime_error("No channel list");
    ptr++;

    //----------------------------------  Loop over channels.
    while (*ptr != '}' && ptr < end && *ptr != ';') {
	string chn[3];
	skip_spaces(ptr, end);
	int chni = 0;
	while (*ptr != ' ' && ptr < end && *ptr != ';' && *ptr != '}') {
	    if ((*ptr == ',' || *ptr == '%') && chni+1 < 3) {
		chni++;
		ptr++;
	    } else {
		chn[chni] += *ptr++;
	    }
	}

	//-------------------------------  Parse channel specifier.
	if (!chn[0].empty()) {
	    if (debug(dlvl_detail)) {
		cout << "Channel specifier: " << chn[0] << "," << chn[1] 
		     << "," << chn[2] << " (" << chni << " pars)" << endl;
	    }
	    chan_type cType = t->getDefType();
	    double rate = 0;
	    for (int i=1; i<=chni; ++i) {
		if (chn[i].empty()) {
		    continue;
		} else if (chn[i][0] >= '0' && chn[i][0] <= '9') {
		    rate = strtod(chn[i].c_str(), 0);
		} else {
		    cType.set(chn[i]);
		}
	    }
	    chan_index inx = Channel_List.find_best(chn[0], cType, rate,
						    t->getStartGPS());
	    if (debug(dlvl_detail)) {
		cout << "Best channel: " << chn[0] << "," << chn[1] 
		     << "," << chn[2] << " (" << chni << " pars)" << endl;
	    }
	    t->add_channel(inx);
	}
    }
    skip_spaces(ptr, end);
    if (*ptr != '}') throw runtime_error("No end brace");
    ptr++;

    //-----------------------------------  Print channel list.
    if (debug(dlvl_detail)) {
	int nChan = t->getNChan();
	if (!nChan) {
	    cout << "No Channels!" << endl;
	} else {
	    cout << "Channel List: ";
	    for (int i=0; i<nChan; ++i) {
	        chan_index j = t->chan(i).index();
		cout << "  " << Channel_List[j].name();
	    }
	    cout << endl;
	}
    }
}

//======================================  Parse channel type
void
parse_stage::parse_chantype(const char* &ptr, const char* end, transaction* t)
{
    skip_spaces(ptr, end);
    const char* p = ptr;
    while (p < end && ((*p >='a' && *p <= 'z') || *p == '-')) p++;
    if (p != ptr) t->setDefType(chan_type(string(ptr, p-ptr)));
    ptr = p;
}

//======================================  Parse command
void 
parse_stage::parse_command(const char* &ptr, const char* end, transaction* t) {
    string cmd;
    skip_spaces(ptr, end);
    while (ptr < end && *ptr != 0) {
	if (*ptr == ' ' || *ptr == '\n' ||
	    *ptr == ';' || *ptr == '{') break;
	cmd += *ptr++;
    }
    t->setCommand(cmd);
    if (debug(dlvl_detail)) {
	cout << "Command:         " 
	     << transaction::getCommandString(t->command()) 
	     << endl;
    }
}

//======================================  Parse epoch
void 
parse_stage::parse_epoch(const char* &ptr, const char* end, transaction* t) {
    //----------------------------------  Get the epoch string.
    string epic_def;
    skip_spaces(ptr, end);
    while (ptr < end && *ptr != 0) {
	if (*ptr == ' ' || *ptr == '\n' ||
	    *ptr == ';' || *ptr == '{') break;
	epic_def += *ptr++;
    }

    //----------------------------------  See if it is a defined name
    if (Epoch_Table.defined(epic_def)) {
	epoch epic = Epoch_Table[epic_def];
	if (epic.none()) throw runtime_error("Invalid epoch");
	t->set_epoch(epic);
	if (debug(dlvl_detail)) cout << "Set epic: " << epic_def << endl;
    }

    //----------------------------------  Parse an epoch definition
    else {
	epoch epic = epoch::parse(epic_def.c_str());
	if (epic.none()) throw runtime_error("Invalid epoch");
	t->set_epoch(epic);
	if (debug(dlvl_detail)) cout << "Set epic: " << epic << endl;
    }
}

//======================================  Parse time list.
int 
parse_stage::parse_times(const char* &ptr, const char* end, 
			 int nMax, long iarg[]) {
    int nArg = 0;
    skip_spaces(ptr, end);
    while (*ptr >='0' && *ptr <='9' && ptr < end) {
	if (nArg >= nMax) throw runtime_error("Too many numeric arguments");
	iarg[nArg++] = strtol(ptr, const_cast<char**>(&ptr), 0);
	skip_spaces(ptr, end);
    }
    if (debug(dlvl_detail)) {
	if (nArg) {
	    cout << "Numeric options: " << iarg[0];
	    for (int i=1; i<nArg; ++i) cout  << "  " << iarg[i];
	    cout << endl;
	} else {
	    cout << "No Numeric Options!" << endl;
	}
    }
    return nArg;
}

//======================================  Thread initialization.
void
parse_stage::init_thread(void) {
    sigignore(SIGPIPE);
}

//======================================  Set thread parameters
void
parse_stage::set_parameters(const stage_pars& p) {
    timeout = p._timeout;
    seda_stage::set_parameters(p);
}

//======================================  Shut down thread pool
void
parse_stage::shut(void) {
    seda_stage::shut();
    cout << "  Quit commands:    " << quit_count << endl;
}
